general_parameters = \
    {'System': '\'VELETA\'',
     'Laser_Pointing_Angle': 0,
     'Molecular_Calc': 0,
     'Latitude_degrees_north': 37.164,
     'Longitude_degrees_east': -3.605,
     'Altitude_meter_asl': 680.0,
     'Call sign': "gr"}

channel_parameters = \
    {'BT0': {'channel_ID': 1363,                    
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'Laser_Shots': 600,
                    'LR_Input': 1,
                    'DAQ_Range': 500.0, },
     'BT1': {'channel_ID': 315,                    
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'Laser_Shots': 600,
                    'LR_Input': 1,
                    'DAQ_Range': 500.0, },
     }