general_parameters = \
    {'System': '\'MULHACEN\'',
     'Laser_Pointing_Angle': 0,
     'Molecular_Calc': 0,
     'Latitude_degrees_north': 37.164,
     'Longitude_degrees_east': -3.605,
     'Altitude_meter_asl': 680.0,
     'Call sign': "gr"}

channel_parameters = \
{'BT0_p45': {'channel_ID': 1055,
         'Background_Low': 75000,
         'Background_High': 105000,
         'LR_Input': 1,         
         'Depolarization_Factor': 0,
         'Pol_Calib_Range_Min': 1000,
         'Pol_Calib_Range_Max': 3000}, 
 'BC0_p45': {'channel_ID': 1057,
         'Background_Low': 75000,
         'Background_High': 105000,
         'LR_Input': 1,         
         'Depolarization_Factor': 0,
         'Pol_Calib_Range_Min': 1000,
         'Pol_Calib_Range_Max': 3000},
 'BT1_p45': {'channel_ID': 1056,
         'Background_Low': 75000,
         'Background_High': 105000,
         'LR_Input': 1,         
         'Depolarization_Factor': 0,
         'Pol_Calib_Range_Min': 1000,
         'Pol_Calib_Range_Max': 3000},
 'BC1_p45': {'channel_ID': 1058,
         'Background_Low': 75000,
         'Background_High': 105000,
         'LR_Input': 1,         
         'Depolarization_Factor': 0,
         'Pol_Calib_Range_Min': 1000,
         'Pol_Calib_Range_Max': 3000},         
 'BT0_m45': {'channel_ID': 1059,
         'Background_Low': 75000,
         'Background_High': 105000,
         'LR_Input': 1,         
         'Depolarization_Factor': 0,
         'Pol_Calib_Range_Min': 1000,
         'Pol_Calib_Range_Max': 3000},
 'BC0_m45': {'channel_ID': 1061,
         'Background_Low': 75000,
         'Background_High': 105000,
         'LR_Input': 1,         
         'Depolarization_Factor': 0,
         'Pol_Calib_Range_Min': 1000,
         'Pol_Calib_Range_Max': 3000},         
 'BT1_m45': {'channel_ID': 1060,
         'Background_Low': 50000,
         'Background_High': 60000,
         'Laser_Shots': 3000,
         'LR_Input': 1,         
         'Pol_Calib_Range_Min': 1000,
         'Pol_Calib_Range_Max': 3000},
 'BC1_m45': {'channel_ID': 1062,
         'Background_Low': 50000,
         'Background_High': 60000,
         'Laser_Shots': 3000,
         'LR_Input': 1,         
         'Pol_Calib_Range_Min': 1000,
         'Pol_Calib_Range_Max': 3000},         
         }