general_parameters = \
    {'System': '\'MULHACEN\'',
     'Laser_Pointing_Angle': 0,
     'Molecular_Calc': 0,
#     'Latitude_degrees_north': 37.164,
#     'Longitude_degrees_east': -3.605,
#     'Altitude_meter_asl': 680.0,
     'Call sign': "gr"}

#BT0: 532xpa - 1284
#BC0: 532xpp - 
#BT1: 532xca - 1283
#BC1: 532xcp
#BT2: 355xta - 1282
#BC2: 355xtp
#BT3: 1064xta - 1045
#BC3: 530xtp - 1048
#BC4: 353xtp - 1047
#BC5: 408xtp

channel_parameters = \
    {'BT0': {'channel_ID': 1284,
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'LR_Input': 1, },
     'BT1': {'channel_ID': 1283,
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'LR_Input': 1, },
     'BT2': {'channel_ID': 1282,
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'LR_Input': 1, },
     'BT3': {'channel_ID': 1045,
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'LR_Input': 1, },               
     'BC0': {'channel_ID': 1094,
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'LR_Input': 1, },               
     'BC1': {'channel_ID': 1093,
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'LR_Input': 1, },               
     'BC2': {'channel_ID': 1090,
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'LR_Input': 1, },               
     'BC4': {'channel_ID': 1047,
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'LR_Input': 1, },               

     }
