general_parameters = \
    {'System': '\'MULHACEN\'',
     'Laser_Pointing_Angle': 0,
     'Molecular_Calc': 0,
#     'Latitude_degrees_north': 37.164,
#     'Longitude_degrees_east': -3.605,
#     'Altitude_meter_asl': 680.0,
     'Call sign': "gr"}

channel_parameters = \
    {'BT0': {'channel_ID': 1284,
#                    'channel_str_ID': 'gr_0532xpa_elPR',
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'LR_Input': 1, },
     'BT1': {'channel_ID': 1283,
#                    'channel_str_ID': 'gr_0532xca_elPT',
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'LR_Input': 1, },
     'BT2': {'channel_ID': 1282,
#                    'channel_str_ID': 'gr_0355xta_elT',
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'LR_Input': 1, },
     'BT3': {'channel_ID': 1045,
#                    'channel_str_ID': 'gr_01064xta',
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'LR_Input': 1, },
     }
